#include <stdio.h>

#include <string.h>
#include <stdint.h>
#include <FL/fl_ask.H>
#include "HidDevice.h"
#include "mapdata.h"

#define MAX_IG_PAGE			20
#define MAX_PV_PAGE			10
#define MAX_CF_PAGE			2

// CDI command
#define CMD_GET_DATA		0x80
#define CMD_ERASE_DATA		0x81
#define CMD_SET_DATA		0x82
#define CMD_GET_STATUS		0x83
#define CMD_GET_DEVIDS		0x84

// 1,000rpm1x̃JEg=666.67(16Mhz Fosc/4)
#define COUNT_OF_1DEG_AT_1000RPM	666.666667

/*
T2CON :rsv.<1>|T2OUTPS<3:0>|TMR2ON|T2CKPS<1:0>
*/
void putbin(uint8_t* p);


HidDevice::HidDevice() {
	msgfunc = 0;
	handle = 0;
	// set default value
	count_of_tune = COUNT_OF_TUNE_2;
	cfg.sys_pickup_degree = PICKUP_DEG_2;
	count_of_1deg_at_1000rpm = COUNT_OF_1DEG_AT_1000RPM;
}

void HidDevice::msg( char* p ) {
	if(msgfunc) {
		msgfunc( p );
	}
}

int HidDevice::open(uint16_t vid, uint16_t pid, wchar_t *s_number) {
	if( handle ) {
		return 0;
	}
	handle = hid_open(vid, pid, s_number);
	if( 0 == handle ) {
		return 0;
	}
	return 1;
}

void HidDevice::close() {
	if( handle ) {
		hid_close( handle );
		handle = 0;
		return;
	}
}

// ---------------------------------------------------------------------------
// checkConfig
// ---------------------------------------------------------------------------
void HidDevice::checkConfig(void) {
	// If the value is wrong, fill with an appropriate value
	if (0 == cfg.rev_limit) {
		cfg.rev_limit = 70;
	}
}


// ---------------------------------------------------------------------------
// readTable
// ---------------------------------------------------------------------------
int HidDevice::readTable(double* pIg, double* pPv) {
	int page, ii, rpm;
	double* pData;


	// --------------------------------------------------------------
	// Read config values.
	// --------------------------------------------------------------
	msg("Reading CONFIG table\n");
	uint8_t* p = (uint8_t*)&cfg;

	for (page = MAX_IG_PAGE + MAX_PV_PAGE; page < MAX_IG_PAGE + MAX_PV_PAGE + MAX_CF_PAGE; page++) {
	txbuf[1] = CMD_GET_DATA;
	txbuf[2] = page; // Set page no to command.
		putbin(&txbuf[0]);

		if (0 > write()) return 0;
		if (0 > read(rxPacket.rxbuf8)) {
			msg("Error occured on reading PV data.\n");
			fl_beep(FL_BEEP_ERROR);
			fl_alert("Error occured on reading PV data.");
			return 0;
		}
		putbin(rxPacket.rxbuf8);

		// Reading 32 values on 1 page.
		for (ii = 0; ii<32; ii++, p++) {
			*p = rxPacket.rxbuf8[4 + ii];
		}
		msg(".");
	}
	checkConfig();
	msg("OK.\n");

	// --------------------------------------------------------------
	// Read ignition page.
	// --------------------------------------------------------------
	msg( "Reading IG table\n" );
	pData = pIg;
	rpm = 0;
	for( page = 0; page < MAX_IG_PAGE; page++ ) {
		txbuf[1] = CMD_GET_DATA;
		txbuf[2] = page; // Set page no to command.

		if( 0 > write() || 0 > read(rxPacket.rxbuf8) ) {
			msg( "Error occured on reading IG data.\n" );
			fl_beep(FL_BEEP_ERROR);
			fl_alert( "Error occured on reading IG data." );
			return 0;
		}

		// read 16 values on 1 page.
		for( ii=0; ii<16; ii++ ) {
			*pData++ = time2deg(rxPacket.rxbuf16[2 + ii], (rpm % 160));
			rpm++;
		}

		puts("TX");
		putbin(txbuf);
		puts("RX");
		putbin(rxPacket.rxbuf8);

		msg( "." );
	}
	msg( "OK.\n" );

	// --------------------------------------------------------------
	// Read power valve values.
	// --------------------------------------------------------------
	pData = pPv;
	msg( "Reading PV table" );
	for( pData = pPv; page<MAX_IG_PAGE+MAX_PV_PAGE; page++ ) {
		txbuf[1] = CMD_GET_DATA;
		txbuf[2] = page; // Set page no to command.
		putbin(&txbuf[0]);

		if( 0 > write() ) return 0;
		if( 0 > read(rxPacket.rxbuf8) ) {
			msg( "Error occured on reading PV data.\n" );
			fl_beep(FL_BEEP_ERROR);
			fl_alert( "Error occured on reading PV data." );
			return 0;
		}
		puts("TX");
		putbin(txbuf);
		puts("RX");
		putbin(rxPacket.rxbuf8);

		// Reading 32 values on 1 page.
		for( ii=0; ii<32; ii++, pData++ ) {
			*pData = rxPacket.rxbuf8[4 + ii];
		}
		msg( "." );
	}
	msg( "OK.\n" );

	return 1;

}

// ---------------------------------------------------------------------------
// writeTable
// ---------------------------------------------------------------------------
int HidDevice::writeTable(double* pIg, double* pPv) {
	int page, ii;
	uint16_t* pTxbuf16 = 0, rpm = 0;


	// --------------------------------------------------------------
	// Write ignition page.
	// --------------------------------------------------------------
	memset( txbuf, 0, sizeof(txbuf) );
	msg( "Writing IG table\n" );
	txbuf[0] = 0x00;

	for( page=0; page<MAX_IG_PAGE; page++ ) {
		txbuf[1] = 0x82;	// WRITE command.
		txbuf[2] = page;

		pTxbuf16 = (uint16_t*)&txbuf[5]; // data started from 5 byte of TX buffer.
		for( ii=0; ii<16; ii++, pTxbuf16++, pIg++, rpm++ ) {
			*pTxbuf16 = deg2time(*pIg, (rpm % 160));
		}
		if(( 0 > write() )||( 0 > read(rxPacket.rxbuf8) )) {
			msg( "Error occured on writing IG data.\n" );
			fl_beep(FL_BEEP_ERROR);
			fl_alert( "Error occured on writing IG data." );
			return 0;
		}
		puts("TX");
		putbin(txbuf);
		puts("RX");
		putbin(rxPacket.rxbuf8);
		if( !compareRxTx() ) {
			msg( "Writing IG data verify failed.\n" );
			fl_beep(FL_BEEP_ERROR);
			fl_alert( "Writing IG data verify failed." );
			return 0;
		}
		msg( "." );
		// verify
	}
	msg( "OK\n" );
	// --------------------------------------------------------------
	// Write power valve values.
	// --------------------------------------------------------------
	rpm = 0;

	for( ; page<MAX_IG_PAGE+MAX_PV_PAGE; page++ ) {
		txbuf[1] = 0x82;	// WRITE command.
		txbuf[2] = page;

		pTxbuf16 = (uint16_t*)&txbuf[5]; // data started from 5 byte of TX buffer.
		for (ii = 0; ii < 32; ii++, pPv++) {
			txbuf[5 + ii] = (uint8_t)*pPv;
		}

		if(( 0 > write() )||( 0 > read(rxPacket.rxbuf8) )) {
			msg( "Error occured on writing PV data.\n" );
			fl_beep(FL_BEEP_ERROR);
			fl_alert( "Error occured on writing PV data." );
			return 0;
		}
		puts("TX");
		putbin(txbuf);
		puts("RX");
		putbin(rxPacket.rxbuf8);

		if( !compareRxTx() ) {
			msg( "Writing PV data verify failed.\n" );
			fl_beep(FL_BEEP_ERROR);
			fl_alert( "Writing PV data verify failed." );
			return 0;
		}
		msg( "." );
		// verify
	}
	msg( "OK\n" );

	// --------------------------------------------------------------
	// Write config values.
	// --------------------------------------------------------------
	uint8_t* p = (uint8_t*)&cfg;

	for( ; page<MAX_IG_PAGE+MAX_PV_PAGE+MAX_CF_PAGE; page++ ) {
		txbuf[1] = 0x82;	// WRITE command.
		txbuf[2] = page;

		pTxbuf16 = (uint16_t*)&txbuf[5]; // data started from 5 byte of TX buffer.
		for( ii=0; ii<32; ii++, p++ ) {
			txbuf[5+ii] = (uint8_t)*p;
		}
		if(( 0 > write() )||( 0 > read( rxPacket.rxbuf8) )) {
			msg( "Error occured on writing config data.\n" );
			fl_beep(FL_BEEP_ERROR);
			fl_alert( "Error occured on writing config data." );
			return 0;
		}
		puts("TX");
		putbin(txbuf);
		puts("RX");
		putbin(rxPacket.rxbuf8);
		if( !compareRxTx() ) {
			msg( "Writing config data verify failed.\n" );
			fl_beep(FL_BEEP_ERROR);
			fl_alert( "Writing config data verify failed." );
			return 0;
		}
		msg( "." );
		// verify
	}
	msg( "OK\n" );

	return 1;

}

// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
int HidDevice::write( const uint8_t* data, size_t length ) {
	if( !handle || (length > HID_MAX_TXBUF_SIZE) ) {
		msg( "Writing fail.\n" );
		return -1;
	}
	txbuf[0] = '\0';

	if( 0 != data ) {
		memset( txbuf, 0, HID_MAX_TXBUF_SIZE); // buf[0] is HID sequence no. always 0
		memcpy( &txbuf[1], data, length );
	}
	int ret = hid_write( handle, txbuf, HID_MAX_TXBUF_SIZE );
	if( ret < 0 ) {
		handle = 0;
		msg( "\nhid_write() Failed.\n" );
	}
	return ret;
}

// ---------------------------------------------------------------------------
// read
// ---------------------------------------------------------------------------
int HidDevice::read( uint8_t* data, size_t length ) {
	if( !handle || (length > HID_MAX_RXBUF_SIZE) ) return -1;

	memset( data,0, length );
	int ret = 0;

	ret = hid_read_timeout( handle, data, length, 1000 );
	if( ret < 0 ) {
		handle = 0;
		msg( "\nhid_rea() Failed.\n" );
	}

	return ret;
}

// ---------------------------------------------------------------------------
// compareRxTx
// ---------------------------------------------------------------------------
bool HidDevice::compareRxTx() {
	for( int ii=4; ii<HID_MAX_RXBUF_SIZE; ii++ ) {
		if( rxPacket.rxbuf8[ii] != txbuf[ii+1] )
			return false;
	}
	return true;
}

// ---------------------------------------------------------------------------
// deg2time
// ---------------------------------------------------------------------------
uint16_t HidDevice::deg2time( double deg, uint16_t rpm ) {
	double pre_degree, pre_count;
	uint16_t prescaler, pre, postscaler;
	uint16_t pr2, t2con;
	uint16_t result, ii;

	if (0 == deg || 0 == rpm) {
		printf("RPM:%3d DEG:%3d PRE:%3d PR2:%3d POST:%3d WRITE:%4X\n", rpm, (int)0, 0, 0, 0, (uint16_t)0);
		return 0;
	}

	// Adjust digit
	deg *=10;
	deg = (int)deg;
	deg /= 10;

	pre_degree = cfg.sys_pickup_degree - deg;
	pre_count = ((pre_degree * count_of_1deg_at_1000rpm * 10) / rpm);
	if (pre_count > count_of_tune) {
		pre_count -= count_of_tune;
	} else {
		pre_count = 0;
	}
	// Timer2 = Prescaler * PR2 * Postscaler 
	// PR2:        {0 ... 255}
	// Prescaler:  {1:1/1:4/1:16}
	// Postscaler: {1:1 ... 1:64}

	// prescaler
	ii = (uint16_t)(pre_count/256);
	if     ((ii/ 1) < 16) {prescaler = 1; pre = 0;}
	else if((ii/ 4) < 16) {prescaler = 4; pre = 1;}
	else if((ii/16) < 16) {prescaler = 16;pre = 2;}
	else if((ii/64) < 16) {prescaler = 64;pre = 3;}
	else {
		printf("*** over flow ***\n");
		printf("RPM:%3d DEG:%3d PRE:%3d PR2:%3d POST:%3d WRITE:%4X\n", rpm, (int)pre_degree, 0, 0, 0, (uint16_t)0);
		return 0;
	}

	// postscaler
	postscaler = (uint16_t)(pre_count/256/prescaler + 1);

	// pr2
	pr2 = (uint16_t)(pre_count / prescaler / postscaler);
	t2con = ((postscaler-1) << 3) + 4  + pre;
	
	result = prescaler * pr2 * postscaler - count_of_tune;
	result = (t2con<<8) + pr2;
	printf("RPM:%3d DEG:%3d PRE:%3d PR2:%3d POST:%3d WRITE:%4X\n", rpm, (int)pre_degree, prescaler, pr2, postscaler, (uint16_t)result);
	return result;
}

// ---------------------------------------------------------------------------
// time2deg
// ---------------------------------------------------------------------------
double HidDevice::time2deg( uint16_t src, uint16_t rpm ) {
	uint8_t t2outps, t2ckps, pre, post, pr2;
	uint32_t count;
	double deg;

	if (0 == src) {
		printf("RPM:%3d DEG:%3d PRE:%3d PR2:%3d POST:%3d READ:%4X\n", rpm, 0, 0, 0, 0, src);
		return 0;
	}

	pr2 = src & 0xff;
	t2outps = src >> 11 & 15;
	t2ckps = (src >> 8) & 3;

	post = t2outps + 1;
	switch( t2ckps ) {
	case 0:	pre = 1;	break;
	case 1: pre = 4;	break;
	case 2: pre = 16;	break;
	case 3: pre = 64;	break;
	default: pre = 64;	 break;// Never in
	}

	count = (pre * pr2 * post) + count_of_tune;
	if (count > 0xFFFF) {
		//return 0;
	}
	deg = cfg.sys_pickup_degree - (((count*rpm) / 10) / count_of_1deg_at_1000rpm);

	// 
	deg *=10;
	deg = (int)deg;
	deg /= 10;

	printf("RPM:%3d DEG:%3d PRE:%3d PR2:%3d POST:%3d READ:%4X\n", rpm, (int)deg, pre, pr2, post, src);

	return  deg;
}


void putbin(uint8_t* p) {
	int ii, jj;
	puts("00-01-02-03-04-05-06-07-08=09-0A-0B-0C-0D-0E-0F");
	for(ii = 0; ii < 4; ii++) {
		for(jj = 0; jj < 16; jj++) {
			printf( "%02X ", (uint8_t)p[ii*16+jj]);
		}
		for(jj = 0; jj < 16; jj++) {
			if( p[ii*16+jj] >= 0x20 && p[ii*16+jj] <=0x7e )
				printf( "%c", p[ii*16+jj]);
			else
				printf( "." );
		}
		

		puts( "" );
	}
}

