#ifdef _DEBUG
#pragma comment( lib, "hidapid.lib" )
#else
#pragma comment( lib, "hidapi.lib" )
#endif

#include <stdint.h>

#include <FL/Fl.H>
#include <FL/Fl_Double_Window.H>
#include <FL/Fl_Native_File_Chooser.H>
#include <FL/fl_ask.H>

#include "flscreen.h"
#include "HidDevice.h"
#include "mapdata.h"

// define
#define WINDOW_TITLE "OSR-CDI v1.2.0"
#define DEFAULT_FILE_NAME "autosaved.cdi"
#define STRING_BUFFER_SIZE 1024

#define IGN_CHART_MAX	35
#define PVS_CHART_MAX	160

// prototype
void repeat_callback(void*);
void msg_cb(char* p);
int dataFileWrite(const char* file_name);
int dataFileRead(const char* file_name);
void option_redraw();
void main_window_cb(Fl_Widget*, void*);

// global
Fl_Double_Window* w;
HidDevice dev;			// CDI device connector
MapData ignMapData;
MapData pvsMapData;
char szDefaultFileName[STRING_BUFFER_SIZE];

// ----------------------------------------------------------------------------
// main
// ----------------------------------------------------------------------------

int main(int argc, char **argv) {
	// -----------------------------------------------------------------
	// Make window
	// -----------------------------------------------------------------
	w = make_window();
	w->callback(main_window_cb);
	w->label(WINDOW_TITLE);
	
	// -----------------------------------------------------------------
	// Setup widgets
	// -----------------------------------------------------------------
	// Init Ignition table & UI
	ignMapData.SetMinMax(0, IGN_CHART_MAX);
	chart_ign->SetMapData(&ignMapData);
	chart_ign->SetChartSize(IGN_CHART_MAX, 5, 5);

	// Init Power valve table & UI
	pvsMapData.SetMinMax(0, PVS_CHART_MAX);
	chart_pvs->SetMapData(&pvsMapData);
	chart_pvs->SetChartSize(PVS_CHART_MAX, 5, 25);

	// Get default path and make default file name.
	wchar_t wstr[STRING_BUFFER_SIZE];
	size_t returnvalue;
	GetTempPath(STRING_BUFFER_SIZE, wstr);
	wcstombs_s(&returnvalue, szDefaultFileName, STRING_BUFFER_SIZE, wstr, STRING_BUFFER_SIZE);
	strcat_s(szDefaultFileName, STRING_BUFFER_SIZE, DEFAULT_FILE_NAME);

	// read default .cdi
	dataFileRead(szDefaultFileName);

	// Set function pointer to message 
	dev.callback(msg_cb);

	// default active chart is ignition map0
	btn_select_ign_map0->value(1);
	chart_ign->SetMapIndex(0);

	tabs_window->value(tab_device_infomation);
	// Now setup is done.
	w->show(argc, argv);
	Fl::add_timeout(0.5, repeat_callback);
	msg->add("Started successfully.\n");
	msg->add("Waiting device connect.\n");


	return(Fl::run());
}

// ----------------------------------------------------------------------------
// Window exit
// ----------------------------------------------------------------------------
static void main_window_cb(Fl_Widget*, void*) {
	dataFileWrite(szDefaultFileName);
	exit(0);
}

// -----------------------------------------------------------------
// Message function pointer
// -----------------------------------------------------------------
static void msg_cb(char* p) {
	msg->add((const char*)p);
#ifdef _DEBUG
	printf("%s", p);
#endif
}


// ----------------------------------------------------------------------------
// repeat callback
// ----------------------------------------------------------------------------
static void repeat_callback(void*) {
	hid_device *handle;
	struct hid_device_info *dev;
	char str[STRING_BUFFER_SIZE] = { 0 };
	usbPacket rxPacket;
	uint8_t txPacket[65] = { 0 };
	size_t returnvalue;

	//return;

	dev = hid_enumerate(DEVICE_VID, DEVICE_PID);
	if (dev) {
		if (dev->manufacturer_string) {
			wcstombs_s(&returnvalue, str, STRING_BUFFER_SIZE, dev->manufacturer_string, STRING_BUFFER_SIZE);
			dev_manufacture->value(str);
		}
		if (dev->product_string) {
			wcstombs_s(&returnvalue, str, STRING_BUFFER_SIZE, dev->product_string, STRING_BUFFER_SIZE);
			dev_product->value(str);
		}
		hid_free_enumeration(dev);
	}
	else {
		dev_manufacture->value("");
		dev_product->value("");
	}


	handle = hid_open(DEVICE_VID, DEVICE_PID, NULL);
	if (handle) {
		txPacket[0] = 0x00;
		txPacket[1] = 0x83; // CMD_GET_STATUS;
		if (hid_write(handle, (unsigned char*)&txPacket, 65) > 0) {
			if (hid_read(handle, (unsigned char*)&rxPacket, 64) > 0) {
				dev_mapsw->color(rxPacket.s.sw1 ? FL_CYAN : FL_MAGENTA);
				dev_mapsw->value(rxPacket.s.sw1 ? "MAP1" : "MAP0");

				dev_quickshifter->color(rxPacket.s.qs_signal ? FL_MAGENTA : FL_CYAN);
				dev_quickshifter->value(rxPacket.s.qs_signal ? "QSSW OFF" : "QSSW ON");
				_itoa_s(rxPacket.s.rpm, str, STRING_BUFFER_SIZE, 10);
				dev_rpm->value(str);
				_itoa_s(rxPacket.s.pv_pot, str, STRING_BUFFER_SIZE, 10);
				dev_powervalve->value(str);
			}
		}

		dev_devid->value("00");
		dev_revid->value("00");

		hid_close(handle);
	}
	else {
		dev_mapsw->value("");
		dev_quickshifter->value("");
		dev_rpm->value("");
		dev_powervalve->value("");
		dev_devid->value("");
		dev_revid->value("");
	}
	Fl::repeat_timeout(0.5, repeat_callback);
}

// ----------------------------------------------------------------------------
// Up setting button
// ----------------------------------------------------------------------------
void cb_up(Fl_Repeat_Button*, void*) {
	if (btn_select_ign_map0->value()|| btn_select_ign_map1->value()) {
		chart_ign->MovePointerY(0.1);
	}
	if (btn_select_pvs_map0->value()|| btn_select_pvs_map1->value()) {
		chart_pvs->MovePointerY(1);
	}
	chart_ign->ReMap(chart_ign->GetPointerX());
	chart_ign->redraw();

	chart_pvs->ReMap(chart_ign->GetPointerX());
	chart_pvs->redraw();
}

// ----------------------------------------------------------------------------
// Down setting button
// ----------------------------------------------------------------------------
void cb_down(Fl_Repeat_Button*, void*){
	if (btn_select_ign_map0->value() || btn_select_ign_map1->value()) {
		chart_ign->MovePointerY(-0.1);
	}
	if (btn_select_pvs_map0->value() || btn_select_pvs_map1->value()) {
		chart_pvs->MovePointerY(-1);
	}
	chart_ign->ReMap(chart_ign->GetPointerX());
	chart_ign->redraw();

	chart_pvs->ReMap(chart_ign->GetPointerX());
	chart_pvs->redraw();
}

// ----------------------------------------------------------------------------
// Left setting button
// ----------------------------------------------------------------------------
void cb_left(Fl_Repeat_Button*, void*){
	chart_ign->MovePointerX(-1);
	chart_ign->redraw();
	chart_pvs->MovePointerX(-1);
	chart_pvs->redraw();
}

// ----------------------------------------------------------------------------
// Right setting button
// ----------------------------------------------------------------------------
void cb_right(Fl_Repeat_Button*, void*){
	chart_ign->MovePointerX(1);
	chart_ign->redraw();
	chart_pvs->MovePointerX(1);
	chart_pvs->redraw();
}

// ----------------------------------------------------------------------------
// File open button down
// ----------------------------------------------------------------------------
void cb_file_open(Fl_Button*, void*) {
	char str[STRING_BUFFER_SIZE] = { 0 };
	Fl_Native_File_Chooser fnfc;
	fnfc.options(Fl_Native_File_Chooser::NO_OPTIONS);
	fnfc.type(Fl_Native_File_Chooser::BROWSE_FILE);
	fnfc.title("Load CDI data");
	fnfc.filter("CDI data\t*.cdi");
	int ret = fnfc.show();

	if (0 == ret) {

		// 2pass file read
		if (dataFileRead(fnfc.filename())&& dataFileRead(fnfc.filename())) {
			chart_ign->redraw();
			chart_pvs->redraw();

			// rename window title
			strcpy_s(str, STRING_BUFFER_SIZE, WINDOW_TITLE);
			strcat_s(str, STRING_BUFFER_SIZE, " - ");
			strcat_s(str, STRING_BUFFER_SIZE, fnfc.filename());
			w->label(str);
			return;
		}
	}
	else if (ret < 0) {
		fl_beep(FL_BEEP_ERROR);
		fl_alert("Cannot open file or write data.");
	}
}

// ----------------------------------------------------------------------------
// File save button down
// ----------------------------------------------------------------------------
void cb_file_save(Fl_Button*, void*) {
	char str[STRING_BUFFER_SIZE] = { 0 };
	Fl_Native_File_Chooser fnfc;
	fnfc.type(Fl_Native_File_Chooser::BROWSE_SAVE_FILE);
	fnfc.title("Save CDI data");
	fnfc.filter("CDI data\t*.cdi");
	fnfc.options(Fl_Native_File_Chooser::SAVEAS_CONFIRM | Fl_Native_File_Chooser::NEW_FOLDER);
	int ret = fnfc.show();

	if (0 == ret) {
		if (dataFileWrite(fnfc.filename())) {
			// rename window title
			strcpy_s(str, STRING_BUFFER_SIZE, WINDOW_TITLE);
			strcat_s(str, STRING_BUFFER_SIZE, " - ");
			strcat_s(str, STRING_BUFFER_SIZE, fnfc.filename());
			w->label(str);
			return;
		}
	}
	else if (ret < 0) {
		fl_beep(FL_BEEP_ERROR);
		fl_alert("Cannot open file or write data.");
	}
}

// ----------------------------------------------------------------------------
// Device read button down
// ----------------------------------------------------------------------------
void cb_device_read(Fl_Button*, void*) {
	char str[STRING_BUFFER_SIZE] = { 0 };

	if (dev.open()) {
		uint8_t current_engine_type;

		current_engine_type = dev.cfg.sys_pulse_per_rotation;

		if (FALSE == dev.readTable(ignMapData.GetPointer(), pvsMapData.GetPointer()))
		{
			fl_alert("Reading device map error.");
		}
		else {
			fl_alert("Reading device map successfully.");
			strcpy_s(str, STRING_BUFFER_SIZE, WINDOW_TITLE);
			strcat_s(str, STRING_BUFFER_SIZE, " - device loaded");
			w->label(str);

		}
		dev.close();

		// Engine type check
		if (current_engine_type != dev.cfg.sys_pulse_per_rotation) {
			fl_beep(FL_BEEP_ERROR);
			fl_alert("WARNNING:Engine Type not matched!\nCheck your config.");
			radio_single_pulse->value(1==dev.cfg.sys_pulse_per_rotation);
			radio_twin_pulse->value(2 == dev.cfg.sys_pulse_per_rotation);
		}

		if (0 == dev.cfg.sys_pickup_degree) {
			fl_beep(FL_BEEP_ERROR);
			fl_alert("WARNNING:Pickup degree is zero!\nCheck your config.");
			dev.cfg.sys_pickup_degree = (1 == dev.cfg.sys_pulse_per_rotation ? PICKUP_DEG_1 : PICKUP_DEG_2);
		}

		// chart draw
		chart_ign->redraw();
		chart_pvs->redraw();

		option_redraw();
		return;
	}
	fl_beep(FL_BEEP_ERROR);
	fl_alert("USB CDI not found.");
}

// ----------------------------------------------------------------------------
// Device write button down
// ----------------------------------------------------------------------------
void cb_device_write(Fl_Button*, void*) {
	msg->clear();

	if (dev.open()) {
		fl_beep(FL_BEEP_QUESTION);
		if (0 == fl_ask("Do you realy want write data to CDI?")) {
			dev.close();
			return;
		}
		if (TRUE == dev.writeTable(ignMapData.GetPointer(), pvsMapData.GetPointer())) {
			if (TRUE == dev.readTable(ignMapData.GetPointer(), pvsMapData.GetPointer())) {
				fl_alert("Writing device map successfully.");
				dev.close();
				return;
			}
		}

		fl_alert("Writing device map error.");
		dev.close();
		return;
	}
	fl_beep(FL_BEEP_ERROR);
	fl_alert("USB CDI not found.");
}

// ----------------------------------------------------------------------------
// .CDI File writing
// ----------------------------------------------------------------------------
int dataFileWrite(const char* file_name) {
	char str[STRING_BUFFER_SIZE] = {};
	int len;
	strncpy_s(str, STRING_BUFFER_SIZE, file_name, STRING_BUFFER_SIZE);
	len = strlen(str);

	if (len > STRING_BUFFER_SIZE) return 0;

	if (len > 4) {
		if (0 != strncmp(&str[len - 4], ".cdi", 4)) {
			strcat_s(str, STRING_BUFFER_SIZE, ".cdi");
		}
	}
	FILE* f;
	fopen_s(&f, str, "w+");
	if (f == NULL) {
		return 0;
	}
	int map, index;
	for (map = 0; map < MAX_MAP_NUM; map++) {
		for (index = 0; index < MAX_DATA_NUM; index++) {
			if (0 > fprintf(f, "IG:%d\t%03d\t%f\n", map, index, ignMapData.Get(map,index))) {
				return 0;
			}
		}
	}
	for (map = 0; map < MAX_MAP_NUM; map++) {
		for (index = 0; index < MAX_DATA_NUM; index++) {
			if (0 > fprintf(f, "PV:%d\t%03d\t%f\n", map, index, pvsMapData.Get(map, index))) {
				return 0;
			}
		}
	}
	// Config
	fprintf(f, "QS-RPM\t%03d\n", dev.cfg.qs_enable_rpm);
	fprintf(f, "QS_ONT\t%03d\n", dev.cfg.qs_sw_on_count);
	fprintf(f, "QS-CUT\t%03d\n", dev.cfg.qs_cut_count);
	fprintf(f, "QS-DIS\t%03d\n", dev.cfg.qs_disable_count);
	fprintf(f, "rev_limit\t%03d\n", dev.cfg.rev_limit);
	// v1.1.1
	fprintf(f, "PulseType\t%03d\n", dev.cfg.sys_pulse_per_rotation);
	fprintf(f, "PickupDeg\t%03d\n", dev.cfg.sys_pickup_degree);
	// v1.1.2
	fprintf(f, "QS-STP\t%03d\n", dev.cfg.qs_use_as_stopsw);
	fprintf(f, "QS-REV\t%03d\n", dev.cfg.qs_reverse_onoff_state);

	fclose(f);
	return 1;
}

// ----------------------------------------------------------------------------
// .CDI File reading
// ----------------------------------------------------------------------------
int dataFileRead(const char* file_name) {
	char str[STRING_BUFFER_SIZE] = {};
	int ndummy;
	double read_data;
	FILE* f;

	fopen_s(&f, file_name, "r+");
	if (f == NULL) {
		return 0;
	}
	int map, index;
	for (map = 0; map < MAX_MAP_NUM; map++) {
		for (index = 0; index < MAX_DATA_NUM; index++) {
			read_data = 0;
			if (EOF == fscanf_s(f, "%s\t%d\t%lf", str, STRING_BUFFER_SIZE, &ndummy, &read_data)) {
				goto FILE_READ_FAIL;
			}
			ignMapData.Set(map, index, read_data);
		}
	}
	for (map = 0; map < MAX_MAP_NUM; map++) {
		for (index = 0; index < MAX_DATA_NUM; index++) {
			if (EOF == fscanf_s(f, "%s\t%d\t%lf", str, STRING_BUFFER_SIZE, &ndummy, &read_data)) {
				goto FILE_READ_FAIL;
			}
			pvsMapData.Set(map, index, read_data);
		}
	}

	// Config
	if (EOF == fscanf_s(f, "%s\t%d\n", str, STRING_BUFFER_SIZE, &ndummy)) goto FILE_READ_FAIL;
	dev.cfg.qs_enable_rpm = (uint8_t)ndummy;
	if (EOF == fscanf_s(f, "%s\t%d\n", str, STRING_BUFFER_SIZE, &ndummy)) goto FILE_READ_FAIL;
	dev.cfg.qs_sw_on_count = (uint8_t)ndummy;
	if (EOF == fscanf_s(f, "%s\t%d\n", str, STRING_BUFFER_SIZE, &ndummy)) goto FILE_READ_FAIL;
	dev.cfg.qs_cut_count = (uint8_t)ndummy;
	if (EOF == fscanf_s(f, "%s\t%d\n", str, STRING_BUFFER_SIZE, &ndummy)) goto FILE_READ_FAIL;
	dev.cfg.qs_disable_count = (uint8_t)ndummy;
	if (EOF == fscanf_s(f, "%s\t%d\n", str, STRING_BUFFER_SIZE, &ndummy)) goto FILE_READ_FAIL;
	dev.cfg.rev_limit = (uint8_t)ndummy;

	// Engine type detection
	if (EOF == fscanf_s(f, "%s\t%d\n", str, STRING_BUFFER_SIZE, &ndummy)) {
		dev.cfg.sys_pulse_per_rotation = 2; // default twin pulse type
	}
	else {
		dev.cfg.sys_pulse_per_rotation = (uint8_t)ndummy;
	}
	
	if (2 != dev.cfg.sys_pulse_per_rotation && 1 != dev.cfg.sys_pulse_per_rotation) {
		dev.cfg.sys_pulse_per_rotation = 2;  // default twin pulse type
	}

	if (2 == dev.cfg.sys_pulse_per_rotation) {
		dev.cfg.sys_pickup_degree = PICKUP_DEG_2;
		dev.count_of_tune = COUNT_OF_TUNE_2;
		dev.cfg.sys_numerator_for_rpm = SYS_NUMERATOR_FOR_RPM_2;
	}
	else {
		dev.cfg.sys_pickup_degree = PICKUP_DEG_1;
		dev.count_of_tune = COUNT_OF_TUNE_1;
		dev.cfg.sys_numerator_for_rpm = SYS_NUMERATOR_FOR_RPM_1;
	}

	// Pickup coil position read
	if (EOF == fscanf_s(f, "%s\t%d\n", str, STRING_BUFFER_SIZE, &ndummy)) {
		dev.cfg.sys_pickup_degree = 36; // Typical twin value
	}
	else {
		dev.cfg.sys_pickup_degree = (uint8_t)ndummy;
	}

	// V1.1.2 Quickshifter option added
	// qs_use_as_stopsw
	if (EOF == fscanf_s(f, "%s\t%d\n", str, STRING_BUFFER_SIZE, &ndummy)) {
		dev.cfg.qs_use_as_stopsw = 0;
	}
	else {
		dev.cfg.qs_use_as_stopsw = (uint8_t)(ndummy?1:0);
	}

	// qs_reverse_onoff_state
	if (EOF == fscanf_s(f, "%s\t%d\n", str, STRING_BUFFER_SIZE, &ndummy)) {
		dev.cfg.qs_reverse_onoff_state = 0; 
	}
	else {
		dev.cfg.qs_reverse_onoff_state = (uint8_t)(ndummy?1:0);
	}

	option_redraw();
	fclose(f);
	return 1;
FILE_READ_FAIL:
	option_redraw();
	fclose(f);
	return 0;
}

// ----------------------------------------------------------------------------
// Redraw option screen
// Copy device data to user interface
// ----------------------------------------------------------------------------
void option_redraw() {
	char str[STRING_BUFFER_SIZE];

	vs_qs_enable_rpm->value((double)dev.cfg.qs_enable_rpm);
	vs_qs_sw_on_count->value((double)dev.cfg.qs_sw_on_count);
	vs_qs_cut_count->value((double)dev.cfg.qs_cut_count);
	vs_qs_disable_count->value((double)dev.cfg.qs_disable_count);
	vs_rev_limit->value((double)dev.cfg.rev_limit);

	if (1 == dev.cfg.sys_pulse_per_rotation) {
		radio_single_pulse->value(1);
		radio_twin_pulse->value(0);
	}
	else {
		radio_single_pulse->value(0);
		radio_twin_pulse->value(1);
	}

	_itoa_s(dev.cfg.sys_pickup_degree, str, STRING_BUFFER_SIZE, 10);
	sys_pickup_degree->value(str);

	btn_qs_reverse_onoff_state->value(dev.cfg.qs_reverse_onoff_state);
	btn_qs_use_as_stopsw->value(dev.cfg.qs_use_as_stopsw);
}

// ----------------------------------------------------------------------------
// Option slider_bar changed
// ----------------------------------------------------------------------------
void cb_option_setting_vs(Fl_Value_Slider*, void*) {
	dev.cfg.qs_enable_rpm = (uint8_t)vs_qs_enable_rpm->value();
	dev.cfg.qs_sw_on_count = (uint8_t)vs_qs_sw_on_count->value();
	dev.cfg.qs_cut_count = (uint8_t)vs_qs_cut_count->value();
	dev.cfg.qs_disable_count = (uint8_t)vs_qs_disable_count->value();
	dev.cfg.rev_limit = (uint8_t)vs_rev_limit->value();
}

// ----------------------------------------------------------------------------
// Option button changed
// ----------------------------------------------------------------------------
void cb_option_setting_btn(Fl_Light_Button*, void*) {
	dev.cfg.qs_reverse_onoff_state = (uint8_t)(btn_qs_reverse_onoff_state->value() ? 1 : 0);
	dev.cfg.qs_use_as_stopsw = (uint8_t)(btn_qs_use_as_stopsw->value() ? 1 : 0);
}

// ----------------------------------------------------------------------------
// Callback of engine type changed
// ----------------------------------------------------------------------------
void cb_engine_type(Fl_Round_Button*, void*) {
	dev.cfg.sys_pulse_per_rotation = radio_single_pulse->value() + (radio_twin_pulse->value() * 2);

	if (2 != dev.cfg.sys_pulse_per_rotation && 1 != dev.cfg.sys_pulse_per_rotation) {
		dev.cfg.sys_pulse_per_rotation = 2;  // default twin pulse type
	}

	if (2 == dev.cfg.sys_pulse_per_rotation) {
		dev.cfg.sys_pickup_degree = PICKUP_DEG_2;
		dev.count_of_tune = COUNT_OF_TUNE_2;
		dev.cfg.sys_numerator_for_rpm = SYS_NUMERATOR_FOR_RPM_2;
	}
	else {
		dev.cfg.sys_pickup_degree = PICKUP_DEG_1;
		dev.count_of_tune = COUNT_OF_TUNE_1;
		dev.cfg.sys_numerator_for_rpm = SYS_NUMERATOR_FOR_RPM_1;
	}
	option_redraw();
}

// ----------------------------------------------------------------------------
// Callback of pickup degree textbox changed
// ----------------------------------------------------------------------------
void cb_sys_pickup_degree(Fl_Int_Input*, void*) {
	dev.cfg.sys_pickup_degree = atoi(sys_pickup_degree->value());
}

// ----------------------------------------------------------------------------
// Callback of selecting map
// ----------------------------------------------------------------------------
void cb_select_map(Fl_Light_Button*, void*) {
	chart_ign->SetMapIndex(SELECTED_NONE);
	chart_pvs->SetMapIndex(SELECTED_NONE);

	if (btn_select_ign_map0->value()) { chart_ign->SetMapIndex(0); }
	if (btn_select_ign_map1->value()) { chart_ign->SetMapIndex(1); }
	if (btn_select_pvs_map0->value()) { chart_pvs->SetMapIndex(0); }
	if (btn_select_pvs_map1->value()) { chart_pvs->SetMapIndex(1); }

	chart_ign->redraw();
	chart_pvs->redraw();
}